#myLog.py
import logging, getpass, time, os


class AKLog(object):
    #class CALog(logging.Logger):
# Reference - 
# http://yhhuang1966.blogspot.com/2018/04/python-logging_24.html
    def __init__(self):
        user=getpass.getuser()
        self.logger=logging.getLogger(user)
        self.logger.setLevel(logging.INFO)
        format='%(asctime)s-%(levelname)s: %(message)s'
        formatter=logging.Formatter(format, datefmt='%Y%m%d-%H%M%S')
        streamhandler=logging.StreamHandler()
        streamhandler.setFormatter(formatter)
        self.logger.addHandler(streamhandler)
        if not os.path.isdir('logs'):
            os.makedirs('logs')
        logfile='./logs/' + user + time.strftime("-%Y-%b") + '.log'
        filehandler=logging.FileHandler(logfile, encoding="utf-8")
        filehandler.setFormatter(formatter)
        self.logger.addHandler(filehandler)
    def set_debug(self):
        self.logger.setLevel(logging.DEBUG)
    def debug(self, msg):
        self.logger.debug(msg)
    def info(self, msg):
        self.logger.info(msg)
    def warning(self, msg):
        self.logger.warning(msg)
    def error(self, msg):
        self.logger.error(msg)
    def critical(self, msg):
        self.logger.critical(msg)
    def log(self, level, msg):
        self.logger.log(level, msg)
    def setLevel(self, level):
        self.logger.setLevel(level)
    def disable(self):
        logging.disable(50)


import xmltodict
from requests.models import Response
from akgoodreads.definitions import Book, Author

class GoodreadsParser:
    def __init__(self, log):
        self.log = log

    def author_id(self, response: Response) -> int:
        """Returns a the goodreads author id from the passed response object
        """
        res = parse_response(response)
        return int(res.get('@id', 0))
    
    def author(self, response: Response) -> Author:
        """Get goodreads author details from author_id response
        """
        res = parse_response(response)
        return _parse_author_info(res)

    def book_ids_from_query(self, response) -> list[int]:
        """Returns a list of book_ids from the xml response of a query
        """
        res = parse_response(response)
        works = res['results']['work']
        return [int(work['best_book']['id']['#text']) for work in works]
    
    def book(self, response) -> Book:
        """Returns the `Book` dataclass parsed from the xml response
        """
        res = parse_response(response)
        return _parse_book_info(res)

    def __repr__(self) -> str:
        return f"GoodreadsParser()"



def parse_response(response: Response):
    res = xmltodict.parse(response.text)['GoodreadsResponse']
    request_status = res.pop('Request')
    assert request_status['authentication'] == 'true'
    assert len(res.keys()) == 1
    return list(res.values())[0]

def _parse_author_info(author: dict) -> Author:
    books = author.get('books').get('book')
    if type(books) == list:
        book_ids = [int(book['id']['#text']) for book in books]
    else:
        book_ids = [int(books['id'])]
    return Author(
            id=int(author.get('id', 0)),
            name=author.get('name', 'NA'),
            link=author.get('link', 'NA'),
            image=author.get('large_image_url',author.get('image_url',author.get('small_image_url','NA'))),
            works=int(author.get('works_count', 0)),
            hometown=author.get('hometown', 'NA'),
            in_goodreads=author.get('goodreads_author') == 'true',
            books=book_ids
        )

def _parse_book_info(book:dict) -> Book:
    authors = book.get('authors').get('author')
 
    if type(authors) == list:
        author_ids = [int(author['id']) for author in authors]
    else:        
        author_ids = [int(authors['id'])]

    return Book(
        id=int_of(book.get('id',0)),
        isbn=int_of(book.get('isbn', 0)),
        isbn13=int_of(book.get('isbn13', 0)),
        title=book.get('title', 'NA'),
        image=book.get('large_image_url',book.get('image_url',book.get('small_image_url','NA'))),
        link=book.get('link', 'NA'),
        pages=int_of(book.get('num_pages',0)),
        publish_year=int_of(book.get('publication_year', 0)),
        rating=float(book.get('average_rating', 0)),
        num_ratings=int_of(book.get('ratings_count', 0)),
        desc=book.get('description', 'NA'),
        authors=author_ids
    )


def int_of(string: str) -> int:
    try:
        return int(string)
    except:
        return 0


import sys
if sys.platform=="win32":
    #pip install keyring
    import keyring


def getpwd(item, username):
    if sys.platform=="win32":
        pwd = keyring.get_password(item, username)
        if not pwd:
            from rich.prompt import Prompt
            print("Password is not saved in keyring.")
            pwd = Prompt.ask(f"Enter the password for {item} corresponding to Username:{username}",password=True)
            choice = input("Would you like to save this pwd to keyring?(Y|n): ")
            if choice.strip().lower() in ['', 'y','n']:
                if choice.strip().lower() != 'n':
                    keyring.set_password(item, username, pwd)
    else:
        from rich.prompt import Prompt
        pwd = Prompt.ask(f"Enter the password for {item} corresponding to Username:{username}",password=True)
    
    return pwd


from akgoodreads.main import *
import pytest

class TestGoodreads:
    client = Goodreads('rpakishore@gmail.com')

    def test__author_from_id(self):
        response = self.client._author_from_id(18541)
        assert response.id == 18541
        assert response.name != "NA"
        assert response.link != "NA"
        assert response.image != "NA"
        assert response.works >= 70
        assert response.hometown == "Cork"
        assert type(response.in_goodreads) == bool
        assert 2581204 in response.books

    def test_author(self):
        response = self.client.author('James A. Michener')
        assert response.id == 7995

    def test_book(self):
        response = self.client.book("Ender's Game", limit=2)
        assert len(response) == 2
        book = response[0]
        assert type(book.id) == int
        assert book.link != "NA"
        assert book.image != "NA"
        assert "Ender’s Game" in book.title

        book = response[1]
        assert type(book.id) == int
        assert book.title != "NA"
        assert book.link != "NA"
        assert book.image != "NA"

    def test__book_from_id(self):
        response = self.client._book_from_id(50)
        assert response.id == 50
        assert response.isbn == 689840926
        assert response.isbn13 == 9780689840920
        assert response.title == "Hatchet (Brian's Saga, #1)"
        assert len(response.image) > 15
        assert response.link == 'https://www.goodreads.com/book/show/50.Hatchet'
        assert response.pages == 208
        assert response.publish_year == 2000
        assert response.rating == 3.76
        assert response.num_ratings == 343233
        assert len(response.desc) > 50
        assert response.authors == [18]
        



from dataclasses import dataclass

@dataclass
class Book:
    id: int
    isbn: int
    isbn13: int
    title: str
    image: str
    link: str
    pages: int
    publish_year: int
    rating: float
    num_ratings: int
    desc: str
    authors: list[int]

@dataclass
class Author:
    id: int
    name: str
    link: str
    image: str
    works: int
    hometown: str
    in_goodreads: bool
    books: list[int]


import requests
import time
from akgoodreads.gr_parser import GoodreadsParser
from akgoodreads.log import AKLog
from akgoodreads.credentials import getpwd
from akgoodreads.definitions import Book, Author

class Goodreads:
    def __init__(self, email:str, sec_bet_req:float=1, debug:bool = False) -> None:
        self.key = getpwd('goodreads', email)
        self.sec_bet_req = sec_bet_req

        self.log = AKLog()
        if debug:
            self.log.set_debug()
        s = requests.Session()

        self._update_last_request()
        s.headers.update({'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36'})
        self.s = s

        self.parser = GoodreadsParser(self.log)

    def __repr__(self) -> str:
        apikey = self.key
        sec_bet_req = self.sec_bet_req
        return f"Goodreads({apikey=}, {sec_bet_req=})"
    
    @property
    def nxt_req(self):
        return self.last_request + self.sec_bet_req
    
    def _update_last_request(self):
        self.last_request = time.time()
        self.log.debug(f'{self.last_request=}')

    def request(self, path:str, params: dict={}):
        if not 'key' in params.keys():
            params['key'] = self.key
        result = gr_req(self.log, f"https://www.goodreads.com/{path}", params, self.s, self.nxt_req)
        self._update_last_request()
        return result

    def author(self, name: str) -> Author:
        #https://www.goodreads.com/api/index#search.authors
        result = self.request(f"api/author_url/{name}")
        author_id = self.parser.author_id(result)
        return self._author_from_id(author_id)

    def _author_from_id(self, id:int) -> Author:
        log = self.log
        log.debug(f"Requsting Author with goodreads ID: {id}")
        result = self.request(f"author/show/{id}")
        return self.parser.author(result)

    def _book_from_id(self, id:int) -> Book:
        log = self.log
        log.debug(f"Requsting Book with goodreads ID: {id}")
        result = self.request(f"book/show/{id}")
        return self.parser.book(result)

    def book(self, name: str, limit:int=None) -> list[Book]:
        """Returns book info from the specified query term. `limit` limits the # of match returned, max 20
        """
        result = self.request('search/index.xml', params={'q':name})
        book_ids = self.parser.book_ids_from_query(result)
        if not limit:
            limit = len(book_ids)-1
        return [self._book_from_id(id) for id in book_ids[:limit]]


def gr_req(log, url: str, params:dict ={}, session = None, next_request: float = None):
    log.debug(f"{next_request=}, {time.time()=}")
    if next_request:
        while next_request - time.time() > 0:
            time.sleep(0.1)
    if session:
        log.debug(f"Sending Sessions request to url:{url} with parameters: {params}")
        req = session.get(url, params=params)
    else:
        headers = {'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36'}
        log.debug(f"Sending standalone request to url:{url} with parameters: {params}")
        req = requests.get(url, params=params, headers=headers)
    req.raise_for_status()
    return req